@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2005-2010 Junjiro R. Okajima
+ * Copyright (C) 2005-2011 Junjiro R. Okajima
  *
  * This program, aufs is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -29,9 +29,11 @@ MODULE_PARM_DESC(debug, "debug print");
 module_param_named(debug, aufs_debug, int, S_IRUGO | S_IWUSR | S_IWGRP);
 
 char *au_plevel = KERN_DEBUG;
-#define dpri(fmt, ...) do { \
-	if (au_debug_test()) \
-		printk("%s" fmt, au_plevel, ##__VA_ARGS__); \
+#define dpri(fmt, ...) do {					\
+	if ((au_plevel						\
+	     && strcmp(au_plevel, KERN_DEBUG))			\
+	    || au_debug_test())					\
+		printk("%s" fmt, au_plevel, ##__VA_ARGS__);	\
 } while (0)
 
 /* ---------------------------------------------------------------------- */
@@ -131,6 +133,16 @@ void au_dpri_inode(struct inode *inode)
 			     iinfo->ii_hinode[0 + bindex].hi_whdentry);
 }
 
+void au_dpri_dalias(struct inode *inode)
+{
+	struct dentry *d;
+
+	spin_lock(&dcache_lock);
+	list_for_each_entry(d, &inode->i_dentry, d_alias)
+		au_dpri_dentry(d);
+	spin_unlock(&dcache_lock);
+}
+
 static int do_pri_dentry(aufs_bindex_t bindex, struct dentry *dentry)
 {
 	struct dentry *wh = NULL;
@@ -191,8 +203,8 @@ static int do_pri_file(aufs_bindex_t bindex, struct file *file)
 	    && file->f_dentry
 	    && au_test_aufs(file->f_dentry->d_sb)
 	    && au_fi(file))
-		snprintf(a, sizeof(a), ", mmapped %d",
-			 !!au_fi(file)->fi_hvmop);
+		snprintf(a, sizeof(a), ", gen %d, mmapped %d",
+			 au_figen(file), !!au_fi(file)->fi_hvmop);
 	dpri("f%d: mode 0x%x, flags 0%o, cnt %ld, v %llu, pos %llu%s\n",
 	     bindex, file->f_mode, file->f_flags, (long)file_count(file),
 	     file->f_version, file->f_pos, a);
@@ -222,7 +234,8 @@ void au_dpri_file(struct file *file)
 	if (!fidir)
 		do_pri_file(finfo->fi_btop, finfo->fi_htop.hf_file);
 	else
-		for (bindex = finfo->fi_btop; bindex <= fidir->fd_bbot;
+		for (bindex = finfo->fi_btop;
+		     bindex >= 0 && bindex <= fidir->fd_bbot;
 		     bindex++) {
 			hfile = fidir->fd_hfile + bindex;
 			do_pri_file(bindex, hfile ? hfile->hf_file : NULL);
@@ -244,15 +257,15 @@ static int do_pri_br(aufs_bindex_t bindex, struct au_branch *br)
 		goto out;
 
 	dpri("s%d: {perm 0x%x, cnt %d, wbr %p}, "
-	     "%s, dev 0x%02x%02x, flags 0x%lx, cnt(BIAS) %d, active %d, "
+	     "%s, dev 0x%02x%02x, flags 0x%lx, cnt %d, active %d, "
 	     "xino %d\n",
 	     bindex, br->br_perm, atomic_read(&br->br_count), br->br_wbr,
 	     au_sbtype(sb), MAJOR(sb->s_dev), MINOR(sb->s_dev),
-	     sb->s_flags, sb->s_count - S_BIAS,
+	     sb->s_flags, sb->s_count,
 	     atomic_read(&sb->s_active), !!br->br_xino.xi_file);
 	return 0;
 
- out:
+out:
 	dpri("s%d: err %ld\n", bindex, PTR_ERR(br));
 	return -1;
 }
@@ -332,24 +345,63 @@ void au_dbg_iattr(struct iattr *ia)
 
 /* ---------------------------------------------------------------------- */
 
+void __au_dbg_verify_dinode(struct dentry *dentry, const char *func, int line)
+{
+	struct inode *h_inode, *inode = dentry->d_inode;
+	struct dentry *h_dentry;
+	aufs_bindex_t bindex, bend, bi;
+
+	if (!inode /* || au_di(dentry)->di_lsc == AuLsc_DI_TMP */)
+		return;
+
+	bend = au_dbend(dentry);
+	bi = au_ibend(inode);
+	if (bi < bend)
+		bend = bi;
+	bindex = au_dbstart(dentry);
+	bi = au_ibstart(inode);
+	if (bi > bindex)
+		bindex = bi;
+
+	for (; bindex <= bend; bindex++) {
+		h_dentry = au_h_dptr(dentry, bindex);
+		if (!h_dentry)
+			continue;
+		h_inode = au_h_iptr(inode, bindex);
+		if (unlikely(h_inode != h_dentry->d_inode)) {
+			int old = au_debug_test();
+			if (!old)
+				au_debug(1);
+			AuDbg("b%d, %s:%d\n", bindex, func, line);
+			AuDbgDentry(dentry);
+			AuDbgInode(inode);
+			if (!old)
+				au_debug(0);
+			BUG();
+		}
+	}
+}
+
 void au_dbg_verify_dir_parent(struct dentry *dentry, unsigned int sigen)
 {
 	struct dentry *parent;
 
 	parent = dget_parent(dentry);
-	AuDebugOn(!S_ISDIR(dentry->d_inode->i_mode)
-		  || IS_ROOT(dentry)
-		  || au_digen(parent) != sigen);
+	AuDebugOn(!S_ISDIR(dentry->d_inode->i_mode));
+	AuDebugOn(IS_ROOT(dentry));
+	AuDebugOn(au_digen_test(parent, sigen));
 	dput(parent);
 }
 
 void au_dbg_verify_nondir_parent(struct dentry *dentry, unsigned int sigen)
 {
 	struct dentry *parent;
+	struct inode *inode;
 
 	parent = dget_parent(dentry);
-	AuDebugOn(S_ISDIR(dentry->d_inode->i_mode)
-		  || au_digen(parent) != sigen);
+	inode = dentry->d_inode;
+	AuDebugOn(inode && S_ISDIR(dentry->d_inode->i_mode));
+	AuDebugOn(au_digen_test(parent, sigen));
 	dput(parent);
 }
 
@@ -362,22 +414,22 @@ void au_dbg_verify_gen(struct dentry *parent, unsigned int sigen)
 
 	err = au_dpages_init(&dpages, GFP_NOFS);
 	AuDebugOn(err);
-	err = au_dcsub_pages_rev(&dpages, parent, /*do_include*/1, NULL, NULL);
+	err = au_dcsub_pages_rev_aufs(&dpages, parent, /*do_include*/1);
 	AuDebugOn(err);
 	for (i = dpages.ndpage - 1; !err && i >= 0; i--) {
 		dpage = dpages.dpages + i;
 		dentries = dpage->dentries;
 		for (j = dpage->ndentry - 1; !err && j >= 0; j--)
-			AuDebugOn(au_digen(dentries[j]) != sigen);
+			AuDebugOn(au_digen_test(dentries[j], sigen));
 	}
 	au_dpages_free(&dpages);
 }
 
 void au_dbg_verify_kthread(void)
 {
-	if (au_test_wkq(current)) {
+	if (current->flags & PF_WQ_WORKER) {
 		au_dbg_blocked();
-		BUG();
+		WARN_ON(1);
 	}
 }
 
